// zlib による圧縮のサンプル
#include	<stdio.h>
#include	<string.h>
#include	<zlib.h>


// バッファサイズ
#define		IN_BUFSIZE		4096	// 入力用
#define		OUT_BUFSIZE		4096	// 出力用


// 入出力用の構造体
typedef		struct Cio
{
	const char	*in_path;	// 入力ファイル名
	const char	*out_path;	// 出力ファイル名
	FILE		*in;		// 入力ファイル
	FILE		*out;		// 出力ファイル

	z_stream	z;	// 圧縮用の構造体

	char	in_buf[ IN_BUFSIZE ];		// 入力用のバッファ
	char	out_buf[ OUT_BUFSIZE ];		// 出力用バッファ
}	Cio, *hCio;


// エラーメッセージを出力する
static	void	Cio_ZError( hCio h, const char *msg )
{
	if ( msg != NULL )
		fprintf( stderr, "%s: ", msg );
	fprintf( stderr, "%s\n", h->z.msg );
}

// 入出力ファイルを開く
// 失敗した場合，in と out を NULL にして -1 を返す
static	int		Cio_Open( hCio h, const char *in_path, const char *out_path )
{
	// 0 で初期化する
	memset( h, 0, sizeof( *h ) );

	// パスを保存する
	h->in_path = in_path;
	h->out_path = out_path;

	if ( ( h->in = fopen( in_path, "rb" ) ) == NULL )
	{
		// 入力ファイルを開くのに失敗
		perror( in_path );
		h->out = NULL;
		return	-1;
	}
	else if ( ( h->out = fopen( out_path, "wb" ) ) == NULL )
	{
		// 出力ファイルを開くのに失敗
		perror( out_path );
		h->in = NULL;
		return	-1;
	}

	return	0;
}

// 入出力ファイルを閉じる
static	void	Cio_Close( hCio h )
{
	if ( h->in != NULL )
		fclose( h->in );
	if ( h->out != NULL )
		fclose( h->out );
}

// 入力バッファを埋める
static	int		Cio_Fill( hCio h )
{
	// 入力ファイルからの読み込み
	h->z.next_in = h->in_buf;
	h->z.avail_in = fread( h->in_buf, 1, IN_BUFSIZE, h->in );
	if ( h->z.avail_in <= 0 )
	{
		if ( ferror( h->in ) )
		{
			// 入力エラー
			perror( h->in_path );
			return	-1;
		}
		// 入力ファイルの終端
		return	0;
	}

	return	0;
}

// 出力バッファを空にする
static	int		Cio_Flush( hCio h )
{
	// 出力ファイルへの書き込み
	if ( fwrite( h->out_buf, OUT_BUFSIZE - h->z.avail_out, 1, h->out ) != 1 )
	{
		// 出力エラー
		perror( h->out_path );
		return	-1;
	}
	h->z.next_out = h->out_buf;
	h->z.avail_out = OUT_BUFSIZE;

	return	0;
}

// 圧縮の準備をする
static	int		Cio_Start( hCio h )
{
	int		ret;

	// メモリ確保はライブラリに任せる
	h->z.zalloc = Z_NULL;
	h->z.zfree = Z_NULL;
	h->z.opaque = Z_NULL;

	// 圧縮用の構造体を初期化する
	if ( ( ret = deflateInit( &h->z, Z_DEFAULT_COMPRESSION ) ) != Z_OK )
	{
		Cio_ZError( h, NULL );
		return	-1;
	}

	// 入出力を初期化する
	h->z.next_in = h->in_buf;
	h->z.avail_in = 0;
	h->z.next_out = h->out_buf;
	h->z.avail_out = OUT_BUFSIZE;

	return	0;
}

// 圧縮の終了処理をする
static	int		Cio_End( hCio h )
{
	int		ret;

	// 圧縮用の構造体を破棄する
	if ( ( ret = deflateEnd( &h->z ) ) != Z_OK )
	{
		Cio_ZError( h, NULL );
		return	-1;
	}

	return	0;
}

// 圧縮のためのループ
static	int		Cio_Loop( hCio h )
{
	int		ret, flush = Z_NO_FLUSH;

	while ( ( ret = deflate( &h->z, flush ) ) != Z_STREAM_END )
	{
		switch ( ret )
		{
		case	Z_OK:			// 問題なし
		case	Z_BUF_ERROR:	// 入力バッファが空か出力バッファが満杯
			// 入力バッファを埋める
			if ( !h->z.avail_in )
			{
				if ( ( ret = Cio_Fill( h ) ) )
					return	-1;

				// 入力ファイルの終端に達したら動作を変更する
				if ( !h->z.avail_in )
					flush = Z_FINISH;
			}
			// 出力バッファを空にする
			if ( !h->z.avail_out )
			{
				if ( ( ret = Cio_Flush( h ) ) )
					return	-1;
			}
			break;
		case	Z_STREAM_ERROR:
			// 入出力が想定外のエラー（例えば，next_in, next_out が NULL）
			Cio_ZError( h, NULL );
			return	-1;
		}
	}

	// 出力バッファの残りを出力ファイルに書き出す
	return	Cio_Flush( h );
}

// 圧縮する
static	int		Cio_Compress( hCio h )
{
	int		ret;

	// 圧縮の準備
	if ( ( ret = Cio_Start( h ) ) )
		return	ret;

	// 圧縮のループ
	ret = Cio_Loop( h );

	// 圧縮の後片付け
	if ( !ret )
		Cio_End( h );
	else
		ret = Cio_End( h );

	return	ret;
}

// in_path を圧縮して out_path に出力する
static	int		Compress( const char *in_path, const char *out_path )
{
	int		ret;
	Cio		io;

	// 入出力ファイルを開く
	if ( ( ret = Cio_Open( &io, in_path, out_path ) ) )
		return	ret;

	// 圧縮する
	ret = Cio_Compress( &io );

	// 入出力ファイルを閉じる
	Cio_Close( &io );

	return	ret;
}

int		main( int argc, char *argv[] )
{
	if ( argc != 3 )
	{
		fprintf( stderr, "Usage: %s Input Output\n", argv[0] );
		return	-1;
	}

	return	Compress( argv[1], argv[2] );
}
