#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<zlib.h>

#include	"zlibio.h"


//
// Macros.
//


// Temporary buffer.
#define		TMP_SIZE	( 1 << 14 )		// 16KB.
#define		MAX_SIZE	( 1 << 30 )		// 1GB.

// Mode.
#define		MODE_READ		( 1 << 0 )	// Mode to read.
#define		MODE_WRITE		( 1 << 1 )	// Mode to write.
#define		MODE_INVERTED	( 1 << 2 )	// Inverted mode.
#define		MODE_EOF		( 1 << 3 )	// EOF flag.


//
// Types.
//


// Function types.
typedef		int		( *inflate_f )( z_streamp, int );
typedef		int		( *fill_f )( hZIO h );

// Functions to fill or flush buffer.
static	int		ZIO_Fill( hZIO h );
static	int		ZIO_Flush( hZIO h );


// Main structure.
typedef		struct ZIO
{
	inflate_f	proc;		// Compress or decompress.
	fill_f		refresh;	// Fill or flush buffer.
	Bytef	**in_buf;	// Pointer to pointer to buffer.
	uInt	*in_n;		// Pointer to number of bytes available.
	Bytef	**ex_buf;	// Pointer to pointer to buffer.
	uInt	*ex_n;		// Pointer to number of bytes available.

	int			level;	// Level.
	int			mode;	// Mode.
	char		*path;	// Path.
	int			valid;	// Stream is initialized or not.
	z_stream	zst;	// Stream.
	FILE		*fp;	// File pointer.
	char		buf[];	// Temporary buffer.
}	ZIO;


//
// Initialize and finish.
//


// Stream.


// Initialize.
static	void	ZIO_InitStream( z_streamp h )
{
	h->zalloc = Z_NULL;
	h->zfree = Z_NULL;
	h->opaque = Z_NULL;
}


// On memory.


// Allocate memory.
static	int		ZMem_Alloc( hZMem h )
{
	int		size;

	h->avail = 0;

	if ( h->buf != NULL )
		return	ZIO_OK;

	size = ( h->size <= 0 ) ? TMP_SIZE : h->size;
	if ( ( h->buf = malloc( size ) ) == NULL )
		return	ZIO_EMEM;

	h->size = size;

	return	ZIO_OK;
}

// Reallocate memory.
static	int		ZMem_Realloc( hZMem h )
{
	int		size;
	void	*buf;

	if ( h->size >= MAX_SIZE )
		return	ZIO_EMEM;

	size = h->size << 1;
	if ( size > MAX_SIZE )
		size = MAX_SIZE;
	if ( ( buf = realloc( h->buf, size ) ) == NULL )
		return	ZIO_EMEM;

	h->buf = buf;
	h->size = size;

	return	ZIO_OK;
}


// Main structure.


// Initialize.
static	void	ZIO_Init( hZIO h )
{
	memset( h, 0, sizeof( *h ) );

	h->level = Z_DEFAULT_COMPRESSION;
	h->path = NULL;
	h->fp = NULL;

	ZIO_InitStream( &h->zst );
}

// Create.
static	hZIO	ZIO_Create( void )
{
	hZIO	h;

	if ( ( h = ( hZIO )malloc( sizeof( *h ) + TMP_SIZE ) ) != NULL )
		ZIO_Init( h );

	return	h;
}

// Finish.
static	void	ZIO_Finish( hZIO h )
{
	int		mode;

	if ( h->path != NULL )
		free( h->path );

	if ( h->valid )
	{
		mode = ( h->mode & MODE_INVERTED ) ? ~h->mode : h->mode;
		if ( mode & MODE_READ )
			inflateEnd( &h->zst );
		else if ( mode & MODE_WRITE )
			deflateEnd( &h->zst );
	}

	if ( h->fp != NULL )
		fclose( h->fp );
}

// Destroy.
static	void	ZIO_Destroy( hZIO h )
{
	ZIO_Finish( h );
	free( h );
}


//
// Zlib.
//


// Convert mode string to integer.
static	int		ZIO_Mode( hZIO h, const char *mode )
{
	while ( *mode != '\0' )
	{
		switch ( *mode )
		{
		case	'r':	// Read.
			h->mode = ( h->mode | MODE_READ ) & ~MODE_WRITE;
			break;
		case	'w':	// Write.
			h->mode = ( h->mode | MODE_WRITE ) & ~MODE_READ;
			break;
		case	'i':	// Inverted.
			h->mode |= MODE_INVERTED;
			break;
		case	'0' ... '9':	// Compression level.
			h->level = *mode - '0';
			break;
		default:
			return	ZIO_EARG;
		}
		mode++;
	}

	return	ZIO_OK;
}

// Setting (compress (true) or decompress (false)).
static	int		ZIO_ConfigCD( hZIO h, int flag )
{
	int		ret;

	if ( flag )
	{
		// Compress.
		ret = deflateInit( &h->zst, h->level );
		h->proc = deflate;
	}
	else
	{
		// Decompress.
		ret = inflateInit( &h->zst );
		h->proc = inflate;
	}

	h->valid = ret == Z_OK;

	return	ret;
}

// Setting (read (true) or write (false)).
static	void	ZIO_ConfigRW( hZIO h, int flag )
{
	if ( flag )
	{
		// Read.
		h->refresh = ZIO_Fill;

		h->in_buf = &h->zst.next_in;
		h->in_n = &h->zst.avail_in;
		h->ex_buf = &h->zst.next_out;
		h->ex_n = &h->zst.avail_out;
	}
	else
	{
		// Write.
		h->refresh = ZIO_Flush;

		h->in_buf = &h->zst.next_out;
		h->in_n = &h->zst.avail_out;
		h->ex_buf = &h->zst.next_in;
		h->ex_n = &h->zst.avail_in;
	}
}

// Ready.
static	int		ZIO_Ready( hZIO h )
{
	int		ret = Z_OK;

	ZIO_ConfigRW( h, h->mode & MODE_READ );
	if ( h->mode & MODE_READ )
		ret = ZIO_ConfigCD( h, h->mode & MODE_INVERTED );
	else if ( h->mode & MODE_WRITE )
	{
		ret = ZIO_ConfigCD( h, ~h->mode & MODE_INVERTED );
		h->zst.avail_out = TMP_SIZE;
	}
	*h->in_buf = h->buf;

	// Check return value.
	switch ( ret )
	{
	case	Z_OK:
		return	ZIO_OK;
	case	Z_MEM_ERROR:
		return	ZIO_EMEM;
	case	Z_STREAM_ERROR:
		return	ZIO_EARG;
	case	Z_VERSION_ERROR:
		return	ZIO_EFMT;
	}

	return	ZIO_EINV;
}

// Fill buffer.
static	int		ZIO_Fill( hZIO h )
{
	int		len;

	// Already finished.
	if ( h->mode & MODE_EOF )
		return	ZIO_EFMT;

	if ( ( len = fread( h->buf, 1, TMP_SIZE, h->fp ) ) < 0 )
		return	ZIO_EIO;

	*h->in_buf = h->buf;
	*h->in_n = len;

	if ( !len )
		h->mode |= MODE_EOF;

	return	ZIO_OK;
}

// Flush buffer.
static	int		ZIO_Flush( hZIO h )
{
	int		len, nbyte;

	nbyte = TMP_SIZE - *h->in_n;
	if ( ( len = fwrite( h->buf, 1, nbyte, h->fp ) ) < nbyte )
		return	ZIO_EIO;

	*h->in_buf = h->buf;
	*h->in_n = TMP_SIZE;

	return	ZIO_OK;
}

// End stream.
static	int		ZIO_WriteEnd( hZIO h )
{
	int		ret;

	while ( ( ret = h->proc( &h->zst, Z_FINISH ) ) != Z_STREAM_END )
	{
		switch ( ret )
		{
		case	Z_OK:
			break;
		case	Z_BUF_ERROR:
			if ( !*h->in_n )
			{
				if ( ( ret = ZIO_Flush( h ) ) != ZIO_OK )
					return	ret;
				break;
			}
		default:
			return	ZIO_EINV;
		}
	}

	return	ZIO_Flush( h );
}

// Compress.
static	int		ZIO_CompressCopy( hZIO h, FILE *fp )
{
	int		len;
	char	*buf;

	if ( ( buf = ( char * )malloc( TMP_SIZE ) ) == NULL )
		return	ZIO_EMEM;

	while ( ( len = fread( buf, 1, TMP_SIZE, fp ) ) > 0 )
	{
		if ( ( len = ZIO_Write( h, buf, len ) ) < 0 )
			break;
	}

	free( buf );

	return	( len >= 0 ) ? ZIO_OK : len;
}

// Decompress.
static	int		ZIO_DecompressCopy( hZIO h, FILE *fp )
{
	int		len;
	char	*buf;

	if ( ( buf = ( char * )malloc( TMP_SIZE ) ) == NULL )
		return	ZIO_EMEM;

	while ( ( len = ZIO_Read( h, buf, TMP_SIZE ) ) > 0 )
	{
		if ( ( len = fwrite( buf, len, 1, fp ) ) < 0 )
			break;
	}

	free( buf );

	return	( len >= 0 ) ? ZIO_OK : len;
}

// Load on memory.
static	int		ZMem_LoadData( hZMem h, hZIO io )
{
	int		ret, size;

	if ( ( ret = ZMem_Alloc( h ) ) != ZIO_OK )
		return	ret;

	while ( ( size = ZIO_Read( io, h->buf + h->avail,
							   h->size - h->avail ) ) > 0 )
	{
		h->avail += size;
		if ( h->avail == h->size )
		{
			if ( ( ret = ZMem_Realloc( h ) ) != ZIO_OK )
				return	ret;
		}
	}

	return	( size < 0 ) ? size : ret;
}


//
// Sub interfaces.
//


// Compress.
static	int		ZIO_CompressSub( const char *src, const char *dst )
{
	int		ret;
	FILE	*fp;
	hZIO	h;

	if ( ( fp = fopen( src, "rb" ) ) == NULL )
		return	ZIO_EIO;
	if ( ( ret = ZIO_Open( &h, dst, "w" ) ) != ZIO_OK )
	{
		fclose( fp );
		return	ret;
	}

	ret = ZIO_CompressCopy( h, fp );

	if ( ret == ZIO_OK )
		ret = ZIO_Close( h );
	else
		ZIO_Close( h );
	fclose( fp );

	return	ret;
}

// Decompress.
static	int		ZIO_DecompressSub( const char *src, const char *dst )
{
	int		ret;
	FILE	*fp;
	hZIO	h;

	if ( ( ret = ZIO_Open( &h, src, "r" ) ) != ZIO_OK )
		return	ret;
	if ( ( fp = fopen( dst, "wb" ) ) == NULL )
	{
		ZIO_Close( h );
		return	ZIO_EIO;
	}

	ret = ZIO_DecompressCopy( h, fp );

	fclose( fp );
	if ( ret == ZIO_OK )
		ret = ZIO_Close( h );
	else
		ZIO_Close( h );

	return	ret;
}

// Open.
static	int		ZIO_OpenSub( hZIO h, const char *path, const char *mode )
{
	int		ret;

	if ( ( ret = ZIO_Mode( h, mode ) ) != ZIO_OK ||
		 ( ret = ZIO_Ready( h ) ) != ZIO_OK )
		return	ret;

	if ( h->mode & MODE_READ )
		h->fp = fopen( path, "rb" );
	else if ( h->mode & MODE_WRITE )
		h->fp = fopen( path, "wb" );

	if ( h->fp == NULL )
		return	ZIO_EIO;

	return	ZIO_OK;
}

// Close.
static	int		ZIO_CloseSub( hZIO h )
{
	int		ret = ZIO_OK;

	if ( h->mode & MODE_WRITE )
		ret = ZIO_WriteEnd( h );

	return	ret;
}

// Read and write.
static	int		ZIO_Proc( hZIO h, void *buf, int n )
{
	int		ret, flush;

	*h->ex_buf = buf;
	*h->ex_n = n;

	while ( *h->ex_n )
	{
		flush = ( h->mode & MODE_EOF ) ? Z_FINISH : Z_NO_FLUSH;
		switch ( ret = h->proc( &h->zst, flush ) )
		{
		case	Z_OK:
			break;
		case	Z_DATA_ERROR:
		case	Z_NEED_DICT:
			return	ZIO_EFMT;
		case	Z_MEM_ERROR:
			return	ZIO_EMEM;
		case	Z_STREAM_END:
			return	n - *h->ex_n;
		case	Z_BUF_ERROR:
			if ( !*h->in_n )
			{
				if ( ( ret = h->refresh( h ) ) != ZIO_OK )
					return	ret;
				break;
			}
			else if ( !*h->ex_n )
				break;
		default:
			return	ZIO_EINV;
		}
	}

	return	n;
}

// Load on memory.
static	int		ZMem_LoadSub( hZMem h, const char *path, const char *mode )
{
	int		ret;
	hZIO	io;

	if ( ( ret = ZIO_Open( &io, path, mode ) ) != ZIO_OK )
		return	ret;

	if ( ( ret = ZMem_LoadData( h, io ) ) != ZIO_OK )
		ZIO_Close( io );
	else
		ret = ZIO_Close( io );

	return	ret;
}

// Save to file.
static	int		ZMem_SaveSub( hZMem h, const char *path, const char *mode )
{
	int		ret, size;
	hZIO	io;

	if ( ( ret = ZIO_Open( &io, path, mode ) ) != ZIO_OK )
		return	ret;

	size = ZIO_Write( io, h->buf, h->avail );

	ret = ZIO_Close( io );

	return	( size < 0 ) ? size : ret;
}


//
// API.
//


// Compress.
int		ZIO_Compress( const char *src, const char *dst )
{
	if ( src == NULL || dst == NULL )
		return	ZIO_EARG;

	return	ZIO_CompressSub( src, dst );
}

// Decompress.
int		ZIO_Decompress( const char *src, const char *dst )
{
	if ( src == NULL || dst == NULL )
		return	ZIO_EARG;

	return	ZIO_DecompressSub( src, dst );
}

// Open (mode is [i][rw][0-6]).
//   i: Inverted mode (read and compress, decompress and write).
//  rw: Read from file or write to file.
// 0-9: Compression level (6: default, 0: faster, 9: smaller).
int		ZIO_Open( hZIO *h, const char *path, const char *mode )
{
	int		ret;

	if ( h == NULL )
		return	ZIO_EHDL;
	*h = NULL;
	if ( path == NULL || mode == NULL )
		return	ZIO_EARG;

	if ( ( *h = ZIO_Create() ) == NULL )
		return	ZIO_EMEM;

	if ( ( ret = ZIO_OpenSub( *h, path, mode ) ) != ZIO_OK )
	{
		ZIO_Destroy( *h );
		*h = NULL;
	}

	return	ret;
}

// Close (flush output buffer).
int		ZIO_Close( hZIO h )
{
	int		ret;

	if ( h == NULL )
		return	ZIO_EHDL;

	ret = ZIO_CloseSub( h );

	ZIO_Destroy( h );

	return	ret;
}

// Read (return number of bytes read).
int		ZIO_Read( hZIO h, void *buf, int n )
{
	if ( h == NULL || ~h->mode & MODE_READ )
		return	ZIO_EHDL;
	else if ( n < 0 || ( n > 0 && buf == NULL ) )
		return	ZIO_EARG;

	return	ZIO_Proc( h, buf, n );
}

// Write (return number of bytes written).
int		ZIO_Write( hZIO h, const void *buf, int n )
{
	if ( h == NULL || ~h->mode & MODE_WRITE )
		return	ZIO_EHDL;
	else if ( n < 0 || ( n > 0 && buf == NULL ) )
		return	ZIO_EARG;

	return	ZIO_Proc( h, ( void * )buf, n );
}

// Load on memory (mode is passed to ZIO_Open()).
// If h->buf is NULL, this function allocates memory by realloc().
int		ZMem_Load( hZMem h, const char *path, const char *mode )
{
	if ( h == NULL )
		return	ZIO_EHDL;
	else if ( path == NULL || mode == NULL )
		return	ZIO_EARG;

	return	ZMem_LoadSub( h, path, mode );
}

// Save to file (mode is passed to ZIO_Open()).
int		ZMem_Save( hZMem h, const char *path, const char *mode )
{
	if ( h == NULL )
		return	ZIO_EHDL;
	else if ( h->buf == NULL || path == NULL || mode == NULL )
		return	ZIO_EARG;

	return	ZMem_SaveSub( h, path, mode );
}
